# Copyright 2020 by Michael Thies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from .package_model import OPCRelationship, OPCTargetMode, RELATIONSHIP_TYPE_CORE_PROPERTIES,\
    RELATIONSHIP_TYPE_THUMBNAIL
from .zip_package import ZipPackageReader, ZipPackageWriter
from .core_properties import OPCCoreProperties

DEFAULT_CORE_PROPERTIES_NAME = "/docProps/core.xml"



# Copyright 2020 by Michael Thies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
""" Tests for the main Reader/Writer functionality of the PyECMA376_2 package"""
import datetime
import os.path
import tempfile
import unittest

import lxml.etree as etree  # type: ignore
import pyecma376_2


class TestZipReader(unittest.TestCase):
    def test_reading_empty_docx(self) -> None:
        file_name = os.path.join(os.path.dirname(__file__), "empty_document.docx")
        reader = pyecma376_2.ZipPackageReader(file_name)

        parts = list(reader.list_parts(True))
        self.assertGreater(len(parts), 0)
        package_rels = list(reader.get_raw_relationships())
        self.assertGreater(len(package_rels), 0)
        document_part = reader.get_related_parts_by_type()[
            'http://schemas.openxmlformats.org/officeDocument/2006/relationships/officeDocument'][0]
        self.assertEqual("/word/document.xml", document_part)
        core_properties_part = reader.get_related_parts_by_type()[
            pyecma376_2.RELATIONSHIP_TYPE_CORE_PROPERTIES][0]
        self.assertEqual(pyecma376_2.DEFAULT_CORE_PROPERTIES_NAME, core_properties_part)
        document_rels = list(reader.get_raw_relationships("/word/document.xml"))
        self.assertGreater(len(document_rels), 0)

        reader.close()

    def test_reading_fragmented(self) -> None:
        file_name = os.path.join(os.path.dirname(__file__), "fragmented.docx")
        reader = pyecma376_2.ZipPackageReader(file_name)

        self.assertIn("/word/document.xml", (n for n, ct in reader.list_parts()))
        with reader.open_part("/word/document.xml") as doc:
            etree.parse(doc)

        reader.close()


class TestZipWriter(unittest.TestCase):
    def test_rewrite_docx(self) -> None:
        fd, new_filename = tempfile.mkstemp(suffix=".docx")
        os.close(fd)
        file_name = os.path.join(os.path.dirname(__file__), "empty_document.docx")
        reader = pyecma376_2.ZipPackageReader(file_name)
        writer = pyecma376_2.ZipPackageWriter(new_filename)

        writer.write_relationships(reader.get_raw_relationships())
        for name, content_type in reader.list_parts():
            with writer.open_part(name, content_type) as w:
                with reader.open_part(name) as r:
                    w.write(r.read())
            relationships = list(reader.get_raw_relationships(name))
            if relationships:
                writer.write_relationships(relationships, name)
        writer.close()
        reader.close()

        with pyecma376_2.ZipPackageReader(new_filename) as reader:
            parts = list(reader.list_parts(True))
            self.assertGreater(len(parts), 0)
            package_rels = list(reader.get_raw_relationships())
            self.assertGreater(len(package_rels), 0)

        os.unlink(new_filename)

    def test_write_example(self) -> None:
        fd, filename = tempfile.mkstemp(suffix=".myx")
        os.close(fd)
        with pyecma376_2.ZipPackageWriter(filename) as writer:
            # Add a part
            with writer.open_part("/example/document.txt", "text/plain") as part:
                part.write("Lorem ipsum dolor sit amet.".encode())

            # Write core properties (meta data)
            # To make those work, we need to add the RELATIONSHIP_TYPE_CORE_PROPERTIES relationship below.
            cp = pyecma376_2.OPCCoreProperties()
            cp.created = datetime.datetime.now()
            with writer.open_part(pyecma376_2.DEFAULT_CORE_PROPERTIES_NAME, "application/xml") as part:
                cp.write_xml(part)

            # Write the packages root relationships
            writer.write_relationships([
                pyecma376_2.OPCRelationship("r1", "http://example.com/my-package-relationship-id", "http://example.com",
                                            pyecma376_2.OPCTargetMode.EXTERNAL),
                pyecma376_2.OPCRelationship("r2", "http://example.com/my-document-rel", "example/document.txt",
                                            pyecma376_2.OPCTargetMode.INTERNAL),
                pyecma376_2.OPCRelationship("r3", pyecma376_2.RELATIONSHIP_TYPE_CORE_PROPERTIES,
                                            pyecma376_2.DEFAULT_CORE_PROPERTIES_NAME,
                                            pyecma376_2.OPCTargetMode.INTERNAL),
            ])
        os.unlink(filename)

    def test_write_fragmented(self) -> None:
        """ Test the writing of interleaved parts. Also test premature writing of ContentTypes (using defaults)"""
        fd, filename = tempfile.mkstemp(suffix=".myx")
        os.close(fd)

        with pyecma376_2.ZipPackageWriter(filename) as writer:
            writer.content_types.default_types['txt'] = "text/plain"
            writer.write_content_types_stream()

            handle = writer.create_fragmented_part("/foo.txt", "text/plain")
            with handle.open() as f:
                f.write(b"Hello, ")
            with writer.open_part("/bar.txt", "text/plain") as f:
                f.write(b"Other part's contents")
            with handle.open(last=True) as f:
                f.write(b"World!")

        with pyecma376_2.ZipPackageReader(filename) as reader:
            with reader.open_part("/foo.txt") as f:
                foo_content = f.read()
            with reader.open_part("/bar.txt") as f:
                bar_content = f.read()
        self.assertEqual(b"Hello, World!", foo_content)
        self.assertEqual(b"Other part's contents", bar_content)

        os.unlink(filename)

    def test_empty_file(self) -> None:
        fd, filename = tempfile.mkstemp(suffix=".myx")
        os.close(fd)

        writer = pyecma376_2.ZipPackageWriter(filename)
        writer.close()

        with pyecma376_2.ZipPackageReader(filename) as reader:
            self.assertEqual([], list(reader.list_parts()))
            self.assertEqual([], list(reader.get_related_parts_by_type()))

            core_properties = reader.get_core_properties()
            self.assertIsNone(core_properties.created)



import datetime
import io
import os.path
import unittest

import pyecma376_2


class CorePropertiesTest(unittest.TestCase):
    def test_parsing_core_properties(self) -> None:
        file_name = os.path.join(os.path.dirname(__file__), "CorePropertiesExample.xml")
        with open(file_name, 'rb') as f:
            cp = pyecma376_2.OPCCoreProperties.from_xml(f)

        self.assertEqual("Reviewed", cp.contentStatus)
        self.assertEqual("OPC Core Properties", cp.title)
        self.assertEqual(datetime.date(2005, 6, 12), cp.created)
        self.assertListEqual([('en-US', "color"), ('en-CA', "colour"), ('fr-FR', "couleur")],
                             cp.keywords)  # type: ignore

    def test_serializing_core_properties(self) -> None:
        cp = pyecma376_2.OPCCoreProperties()
        cp.contentStatus = "Reviewed"
        cp.title = "OPC Core Properties"
        cp.created = datetime.date(2005, 6, 12)
        cp.keywords = [('en-US', "color"), ('en-CA', "colour"), ('fr-FR', "couleur")]

        f = io.BytesIO()
        cp.write_xml(f)
        f.seek(0)

        cp2 = pyecma376_2.OPCCoreProperties.from_xml(f)

        self.assertEqual("Reviewed", cp2.contentStatus)
        self.assertEqual("OPC Core Properties", cp2.title)
        self.assertEqual(datetime.date(2005, 6, 12), cp2.created)
        self.assertListEqual([('en-US', "color"), ('en-CA', "colour"), ('fr-FR', "couleur")],
                             cp2.keywords)  # type: ignore

    def test_read_empty_doc_properties(self) -> None:
        file_name = os.path.join(os.path.dirname(__file__), "empty_document.docx")
        with pyecma376_2.ZipPackageReader(file_name) as reader:
            core_properties = reader.get_core_properties()
        self.assertEqual("Michael Thies", core_properties.creator)
        self.assertEqual(2020, core_properties.created.year)



# Copyright 2020-2021 by Michael Thies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
"""
This module provides a Python model of the OPC Core Properties (according to ECMA 376-2, section 11.). All available
properties (meta attributes) of a package are represented by an object of the `OPCCoreProperties` class. Additionally,
this module defines some helper functions to parse or serialize OPCCoreProperties objects from/into XML documents.
"""

import datetime
import re
from typing import Optional, List, Tuple, IO, Callable, Any, Union

import lxml.etree as etree  # type: ignore

XML_NAMESPACE = "{http://schemas.openxmlformats.org/package/2006/metadata/core-properties}"
XML_NAMESPACE_DC = "{http://purl.org/dc/elements/1.1/}"
XML_NAMESPACE_DCTERMS = "{http://purl.org/dc/terms/}"
XML_NAMESPACE_XSI = "{http://www.w3.org/2001/XMLSchema-instance}"
XML_NAMESPACE_XML = "{http://www.w3.org/XML/1998/namespace}"


def _extract_str(el: etree.Element) -> str:
    """
    Helper function to extract the contained text from an XML tag. To be used as extractor function in OPCCoreProperties
    XML for all standard string fields.
    """
    return el.text


def _to_xml_tag(tag: str, content: str) -> etree.Element:
    """
    Helper function to create an XML tag with a simple text content. To be used as a serializer for all string
    attributes in OPCCoreProperties.
    """
    result = etree.Element(tag)
    result.text = content
    return result


DATETIME_RE = re.compile(r'(\d\d\d\d)(-(\d\d)(-(\d\d)(T(\d\d)(:(\d\d)(:(\d\d)(\.\d+)?([+\-](\d\d):(\d\d)|Z)?)?)?)?)?)?')


def _extract_datetime(el: etree.Element) -> Optional[Union[datetime.datetime, datetime.date]]:
    """
    Extract date or datetime value from W3CDTF encoding (https://www.w3.org/TR/NOTE-datetime), contained in the text of
    an XML tag. To be used as a serializer function for the date attributes of OPCCoreProperties.
    """
    match = DATETIME_RE.match(el.text)
    if not match:
        # TODO log warning?
        return None
    if match[6]:
        return datetime.datetime(
            int(match[1]),
            int(match[3]) if match[3] else 0,
            int(match[5]) if match[5] else 0,
            int(match[7]) if match[7] else 0,
            int(match[9]) if match[9] else 0,
            int(match[11]) if match[11] else 0,
            int(float(match[12]) * 1e6) if match[12] else 0,
            _parse_xsd_date_tzinfo(match[13])
        )
    else:
        return datetime.date(
            int(match[1]),
            int(match[3]) if match[3] else 0,
            int(match[5]) if match[5] else 0,
        )


def _parse_xsd_date_tzinfo(value: str) -> Optional[datetime.tzinfo]:
    """ Helper function for `_extract_datetime()` to parse tzinfo part of a W3CDTF string. """
    if not value:
        return None
    if value == "Z":
        return datetime.timezone.utc
    return datetime.timezone(datetime.timedelta(hours=int(value[1:3]), minutes=int(value[4:6]))
                             * (-1 if value[0] == '-' else 1))


def _datetime_to_xml(tag: str, content: Union[datetime.date, datetime.datetime]) -> etree.Element:
    """
    Serialize date or datetime value into an XML tag in W3CDTF encoding (https://www.w3.org/TR/NOTE-datetime). To be
    used as a serializer function for the date attributes of OPCCoreProperties.
    """
    text: str
    if isinstance(content, (datetime.date, datetime.datetime)):
        text = content.isoformat()
    else:
        # Allow string pass through
        text = content  # type: ignore
    return _to_xml_tag(tag, text)


def _extract_keywords(el: etree.Element) -> List[Tuple[Optional[str], str]]:
    """
    Extract OPC Core Properties keywords list from a complex tag. To be used as a extractor function for the
    OPCCoreProperties.keywords attribute.
    """
    default_lang = el.attrib.get(XML_NAMESPACE_XML + "lang", None)
    result = []
    if el.text is not None and el.text.strip():
        result.append((default_lang, el.text.strip()))
    for key_tag in el:
        result.append((key_tag.attrib.get(XML_NAMESPACE_XML + "lang", default_lang), key_tag.text.strip()))
        if key_tag.tail is not None and key_tag.tail.strip():
            result.append((default_lang, key_tag.tail.strip()))
    return result


def _keywords_to_xml(tag: str, content: List[Tuple[Optional[str], str]]) -> etree.Element:
    """
    Serialize OPC Core Properties keywords list into a complex XML tag. To be used as a serializer function for the
    OPCCoreProperties.keywords attribute.
    """
    result = etree.Element(tag)
    for lang, keyword in content:
        key_tag = etree.Element(XML_NAMESPACE + "value")
        key_tag.text = keyword
        if lang:
            key_tag.attrib[XML_NAMESPACE_XML + 'lang'] = lang
        result.append(key_tag)
    return result


class OPCCoreProperties:
    """
    Class to represent the Core Properties Part, containing the meta data of an OPC package.

    The Core Properties are stored as an XML part in the package and referenced via a reference of type
    http://schemas.openxmlformats.org/.../core-properties from the package root.
    """
    def __init__(self):
        # Dublin Core does not define concrete Types for its attributes. However, we expect 'created', 'modified' and
        # 'lastPrinted' to be datetime-ish values in the dcterms:W3CDTF format (https://www.w3.org/TR/NOTE-datetime).

        self.category: Optional[str] = None  # A categorization of the content of this package. (e.g. "Letter")
        self.contentStatus: Optional[str] = None  # The status of the content. (e.g. "Draft", "Final")
        self.created: Optional[Union[datetime.date, datetime.datetime]] = None  # Date of creation of the resource.
        self.creator: Optional[str] = None  # An entity primarily responsible for making the content of the resource.
        self.description: Optional[str] = None  # An explanation of the content of the resource.
        self.identifier: Optional[str] = None  # An unambiguous reference to the resource within a given context.
        # A list of keywords to support searching and indexing. Each Keyword may or may not have its language specified
        # (using XML lang strings).
        self.keywords: Optional[List[Tuple[Optional[str], str]]] = None
        self.language: Optional[str] = None  # The language of the intellectual content of the resource (see RFC 3066).
        self.lastModifiedBy: Optional[str] = None  # The user who performed the last modification.
        # The date and time of the last printing.
        self.lastPrinted: Optional[Union[datetime.date, datetime.datetime]] = None
        # Date on which the resource was changed.
        self.modified: Optional[Union[datetime.date, datetime.datetime]] = None
        self.revision: Optional[str] = None  # The revision number.
        self.subject: Optional[str] = None  # The topic of the content of the resource.
        self.title: Optional[str] = None  # The name given to the resource.
        self.version: Optional[str] = None  # The version number. This value is set by the user or by the application.

    # Yay, some dynamic coding. This tuple defines all attributes of the object as well as the corresponding XML tag
    # and the transformation functions for converting the XML string value into the correct python type and vice versa.
    ATTRIBUTES: Tuple[Tuple[str, str, Callable[[etree.Element], Any], Callable[[str, Any], etree.Element]], ...] = (
        ('category', XML_NAMESPACE + 'category', _extract_str, _to_xml_tag),
        ('contentStatus', XML_NAMESPACE + 'contentStatus', _extract_str, _to_xml_tag),
        ('created', XML_NAMESPACE_DCTERMS + 'created', _extract_datetime, _datetime_to_xml),
        ('creator', XML_NAMESPACE_DC + 'creator', _extract_str, _to_xml_tag),
        ('description', XML_NAMESPACE_DC + 'description', _extract_str, _to_xml_tag),
        ('identifier', XML_NAMESPACE_DC + 'identifier', _extract_str, _to_xml_tag),
        ('keywords', XML_NAMESPACE + 'keywords', _extract_keywords, _keywords_to_xml),
        ('language', XML_NAMESPACE_DC + 'language', _extract_str, _to_xml_tag),
        ('lastModifiedBy', XML_NAMESPACE + 'lastModifiedBy', _extract_str, _to_xml_tag),
        ('lastPrinted', XML_NAMESPACE + 'lastPrinted', _extract_datetime, _datetime_to_xml),
        ('modified', XML_NAMESPACE_DCTERMS + 'modified', _extract_datetime, _datetime_to_xml),
        ('revision', XML_NAMESPACE + 'revision', _extract_str, _to_xml_tag),
        ('subject', XML_NAMESPACE_DC + 'subject', _extract_str, _to_xml_tag),
        ('title', XML_NAMESPACE_DC + 'title', _extract_str, _to_xml_tag),
        ('version', XML_NAMESPACE + 'version', _extract_str, _to_xml_tag),
    )
    ATTRIBUTES_BY_TAG = {a[1]: a for a in ATTRIBUTES}
    NSMAP = {
        None: XML_NAMESPACE[1:-1],
        'dc': XML_NAMESPACE_DC[1:-1],
        'dcterms': XML_NAMESPACE_DCTERMS[1:-1],
        'xml': XML_NAMESPACE_XML[1:-1],
        'xsi': XML_NAMESPACE_XSI[1:-1],
    }

    def write_xml(self, file: IO[bytes]) -> None:
        """
        Serialize and write these Core Properties into an XML file or package Part, according to the ECMA standard.
        """
        with etree.xmlfile(file, encoding="UTF-8") as xf:
            xf.write_declaration()
            with xf.element(XML_NAMESPACE + "coreProperties", nsmap=self.NSMAP):
                for attr, tag, extractor, serializer in self.ATTRIBUTES:
                    value = getattr(self, attr)
                    if value is not None:
                        xf.write(serializer(tag, value))

    @classmethod
    def from_xml(cls, file: IO[bytes]) -> "OPCCoreProperties":
        """
        Read and parse Core Properties from an XML file or package Part, according to the ECMA standard.
        """
        result = cls()
        for _event, elem in etree.iterparse(file):
            if elem.tag in cls.ATTRIBUTES_BY_TAG:
                attr, tag, extractor, serializer = cls.ATTRIBUTES_BY_TAG[elem.tag]
                setattr(result, attr, extractor(elem))
                elem.clear()
        return result
